from __future__ import annotations

import os
from pathlib import Path

import re
import yaml
from pydantic import BaseModel
from termcolor import colored

n_indent = 4
indent = ' '*n_indent


def debug_msg(label, msg, colour):
    label = f'  [{colored(label, colour)}]'
    msg = colored(msg, colour)
    print(f'{label}: {msg}')


def to_snake_case(camel_case_str: str) -> str:
    """
    Convert a camel case string to a snake case string

    Args:
        camel_case_str (str): input camel case string

    Returns:
        str: generated snake case string
    """
    a = re.compile('((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))')
    cleaned_str = "".join(camel_case_str.split())
    snake_case = a.sub(r'_\1', cleaned_str).lower()
    return snake_case.replace('__', '_')


def indent_str(s: str, n: int = n_indent) -> str:
    """
    Generate a string that's indented by some number of spaces.

    Args:
        s (str): the input string. Can be a multiline string (contains '\n').
        n (int): in number of spaces to indent. Defaults to 4.

    Returns:
        str: the indented version of the string.

    """
    _indent = ' ' * n
    return "\n".join(map(lambda x: f"{_indent}{x}", s.split('\n')))


class CustomDumper(yaml.Dumper):

    def represent_data(self, data):
        if isinstance(data, Path):
            data = str(data)

        try:
            return super().represent_data(data)
        except Exception:
            return self.represent_scalar('tag:yaml.org,2002:str', str(data))


class BaseConfig(BaseModel):

    @classmethod
    def load_cfg(cls, fpath: str | Path):
        with open(fpath, "r") as f:
            return cls.model_validate_json(f.read())

    def save_cfg(self, fpath: str | Path):
        fpath = Path(fpath)
        fpath.parent.mkdir(parents=True, exist_ok=True)
        with open(fpath, "w") as f:
            jstr = self.model_dump_json(indent=2)
            f.write(jstr)

    def __repr__(self):
        return yaml.dump(self.model_dump(), Dumper=CustomDumper)

    def __str__(self):
        return yaml.dump(self.model_dump(), Dumper=CustomDumper)

# If we use PPB in fbn-ci we need to use deployment specific
# locations for data storage to avoid API servers conflicting with each other
provided_path = os.getenv("CANDELA_HOME")
CANDELA_DIR = Path.home() / ".candela" if provided_path is None else Path(provided_path).resolve()
