from __future__ import annotations

import json
from typing import Dict

from fastapi.responses import JSONResponse
from json import JSONDecodeError
from requests import Response
from aiohttp import ClientResponse


class CdError(Exception):

    def __init__(self, details):
        self.details = details
        super().__init__()

    def json_response(self, api_path: str) -> JSONResponse:
        raise NotImplementedError()

    def __str__(self):
        return f'{type(self).__name__}: {self.details}'


class CandelaApiError(CdError):

    def __init__(self, status_code, url, method, reason, details):
        self.status_code = status_code
        self.url = url
        self.method = method
        self.reason = reason
        super().__init__(details)

    @classmethod
    async def from_aiohttp_response(cls, response: ClientResponse) -> CandelaApiError:
        try:
            details = await response.json()
        except JSONDecodeError:
            details = await response.text()
        return CandelaApiError(
            response.status,
            response.url,
            response.method,
            response.reason,
            details
        )

    @classmethod
    def from_requests_response(cls, response: Response) -> CandelaApiError:
        try:
            details = response.json()
        except JSONDecodeError:
            details = response.text
        return CandelaApiError(
            response.status_code,
            response.url,
            response.request.method,
            response.reason,
            details
        )

    def json_response(self, api_path: str) -> JSONResponse:
        return JSONResponse(
            status_code=self.status_code,
            content={
                "path": api_path,
                **self.details
            }
        )

    def __str__(self):
        return f"{self.method} {self.url} failed ({self.status_code}: {self.reason})" \
               f"\nDetails: {json.dumps(self.details, indent=2)}"
