from __future__ import annotations

from pydantic import BaseModel, Field
from pathlib import Path

from candela_kit.common.common import BaseConfig, CANDELA_DIR


class UserProfile(BaseModel):
    domain: str
    user_id: str
    api_url: str
    access_token: str
    default_model: str | None
    default_slot_type: str | None = None
    description: str | None = None

    @classmethod
    def local(cls) -> UserProfile:
        return UserProfile(
            domain="local",
            user_id="local",
            api_url="http://127.0.0.1:8933",
            access_token="local",
            default_model="llama3-8b-q6k",
            default_slot_type='local',
            description="The local user",
        )

    def censor_token(self) -> UserProfile:
        profile = self.model_copy()

        def censor(x):
            if x == "NA":
                return x
            n = len(x) - 10
            return x[:4] + "-" * n + x[-6:]

        profile.access_token = censor(self.access_token)

        return profile


class UserProfiles(BaseConfig):
    current: str = "local"
    profiles: dict[str, UserProfile] = Field(default_factory=lambda: {"local": UserProfile.local()})

    def get(self, profile_name: str | None = None) -> UserProfile:
        if profile_name is None:
            return self.profiles[self.current]
        self._assert_in(profile_name)
        return self.profiles[profile_name]

    def add(self, profile_name: str, profile: UserProfile, as_current: bool = False) -> UserProfiles:
        if not isinstance(profile_name, str):
            raise TypeError("Profile name must be a string.")
        if not isinstance(profile, UserProfile):
            raise TypeError("Profile object must be a UserProfile instance.")

        self.profiles[profile_name] = profile
        if as_current:
            self.current = profile_name
        return self

    def set_current(self, profile_name: str) -> UserProfiles:
        self._assert_in(profile_name)
        self.current = profile_name
        return self

    def delete(self, profile_name: str) -> UserProfiles:
        self._assert_in(profile_name)
        del self.profiles[profile_name]
        return self

    def has_profile(self, profile_name) -> bool:
        return profile_name in self.profiles

    def _assert_in(self, profile_name: str):
        if not self.has_profile(profile_name):
            raise ValueError(f'There is no profile called "{profile_name}".')

    @classmethod
    def load_cfg(cls, fpath: str | Path = None):
        if fpath is None:
            fpath = CANDELA_DIR / 'user_profiles.json'
        else:
            fpath = Path(fpath)
        if not fpath.exists():
            default = UserProfiles()
            default.save_cfg(fpath)
        return super().load_cfg(fpath)

    def save_cfg(self, fpath: str | Path = None):
        if fpath is None:
            fpath = CANDELA_DIR / 'user_profiles.json'
        else:
            fpath = Path(fpath)
        return super().save_cfg(fpath)


def profiles() -> UserProfiles:
    return UserProfiles.load_cfg()
