from abc import ABC, abstractmethod
from collections.abc import Sized
from time import sleep
import pandas as pd
from typing import Iterator, Literal, Iterable
import io
from tabulate import tabulate
import json

from termcolor import colored

from candela_kit.common.spinner import Spinner
from candela_kit.common.common import debug_msg, indent_str
from candela_kit.client import Client
from candela_kit.models import ObjectId, Event, ObjectMetadata, SlotState, ToolModule


def event_print(evt):
    if evt.event_type == 'user':
        print(colored('User', 'blue') + ':')
        print('  ' + evt.content)
        print(colored('Assistant', 'blue') + ':')

    elif evt.event_type == 'chat':
        print(evt.content, end='', flush=True)

    elif evt.event_type == 'intent':
        chunk = evt.content.replace('\n', '\n    ')
        print(colored(chunk, 'green'), end='', flush=True)

    elif evt.event_type == 'info':
        debug_msg('INFO', evt.content, 'cyan')

    elif evt.event_type == 'wait':
        debug_msg('WAIT', evt.content, 'magenta')

    elif evt.event_type == 'sys':
        debug_msg('SYS', evt.content, 'light_grey')

    elif evt.event_type == 'result':
        d = json.loads(evt.content)

        debug_msg('RESULT', d['content_type'], 'green')

        if d['content_type'] == 'table':
            df = pd.read_csv(io.StringIO(d['content']['data']))
            content_str = tabulate(df, headers=df.columns, tablefmt='rounded_outline')
        else:
            content_str = json.dumps(d['content'], indent=2)

        print(indent_str(content_str, n=4))


class SessionContext(ABC):

    def __init__(self, client: Client, slot_type: str, model_id: ObjectId, tool_module_override: ToolModule):
        self.client = client
        self.slot_type = slot_type
        self.model_id = model_id
        self.tool_module_override = tool_module_override
        self.session_meta = None

    @abstractmethod
    def _start_session(self) -> ObjectMetadata:
        pass

    def __enter__(self):
        with Spinner('Assigning slot') as spinner:
            spinner.section(f'Starting {type(self).__name__} context')

            if not self.client.has_slot():
                self.client.assign_slot(self.slot_type)

            meta = self.client.get_slot_metadata()
            spinner.info(f'User assigned to slot {meta.slot_id}')

            state = self.client.get_slot_state()
            while state != SlotState.HOST_READY:
                spinner.message = f'Waiting for {meta.slot_id} to be ready (status = {state.name})'
                sleep(1)
                state = self.client.get_slot_state()
            spinner.done = 'Slot is ready'

        with Spinner('Starting session') as spinner:
            self.client.stop_session()
            self.session_meta = self._start_session()
            spinner.done = f'Session {self.session_meta.obj_id.identifier} is ready'

        print()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # stop session
        self.client.stop_session()


class RunAgent(SessionContext, ABC):

    max_len = 80

    def _submit_prompt(self, prompt: str):
        return self.client.submit_prompt_to_agent(prompt, self.session_meta.obj_id.identifier)

    @staticmethod
    def _user_stream_handler(stream: Iterator[Event]) -> Event | None:
        print('\n' + colored('Assistant', 'blue') + ':')
        return next(stream, None)

    def _chat_stream_handler(self, evt, stream: Iterator[Event]) -> Event | None:

        # remove the whitespace
        while evt.content.strip() == '':
            evt = next(stream, None)
        if evt is None:
            return evt

        line_len = 0
        evt.content = '  ' + evt.content.lstrip()

        while evt is not None and evt.event_type == 'chat':

            chunk = evt.content
            chunk = chunk.replace('\n', '\n  ')

            if '\n' in chunk:
                line_len = len(chunk.split('\n')[-1])
            else:
                line_len += len(chunk)

            if line_len > self.max_len:

                parts = chunk.split(' ')

                to_join = [parts[0], '\n ']
                if len(parts) > 1:
                    to_join += parts[1:]
                chunk = ' '.join(to_join)

                line_len = len(chunk) - len(parts[0]) - 1

            print(chunk, end='', flush=True)

            evt = next(stream, None)

        print()
        return evt

    @staticmethod
    def _intent_stream_handler(evt, stream: Iterator[Event], mode: Literal['quiet', 'debug', 'developer']) -> Event | None:

        if mode == 'quiet':
            with Spinner('Thinking'):
                while evt is not None and evt.event_type == 'intent':
                    evt = next(stream, None)
        else:
            evt.content = evt.content.lstrip()
            print(f'  [{colored("INTENT", "green")}]:')
            print('    ', end='')
            while evt is not None and evt.event_type == 'intent':
                event_print(evt)
                evt = next(stream, None)

        return evt

    @staticmethod
    def _wait_stream_handler(evt, stream: Iterator[Event], mode: Literal['quiet', 'debug', 'developer']) -> Event | None:
        if mode == 'developer':
            while evt is not None and evt.event_type == 'wait':
                event_print(evt)
                evt = next(stream, None)
        else:
            with Spinner('Using tool'):
                while evt is not None and evt.event_type == 'wait':
                    evt = next(stream, None)
        return evt

    @staticmethod
    def _sys_stream_handler(evt, stream: Iterator[Event], mode: Literal['quiet', 'debug', 'developer']) -> Event | None:
        if mode == 'developer':
            event_print(evt)
        return next(stream, None)

    @staticmethod
    def _info_stream_handler(evt, stream: Iterator[Event], mode: Literal['quiet', 'debug', 'developer']) -> Event | None:
        if mode == 'developer':
            while evt is not None and evt.event_type == 'info':
                event_print(evt)
                evt = next(stream, None)
        else:
            with Spinner('Thinking'):
                while evt is not None and evt.event_type == 'info':
                    evt = next(stream, None)
        return evt

    @staticmethod
    def _result_stream_handler(evt, stream: Iterator[Event]) -> Event | None:
        event_print(evt)
        return next(stream, None)

    def chat(self, mode: Literal['quiet', 'debug', 'developer'] = 'developer'):

        while True:
            print(colored('User', 'blue') + ':')
            message = input('  ')
            if message == 'quit()':
                print('  User quit the chat session')
                self.client.stop_session()
                return

            stream = self._submit_prompt(message)

            evt = next(stream)
            while evt:
                if evt.event_type == 'chat':
                    evt = self._chat_stream_handler(evt, stream)
                elif evt.event_type == 'intent':
                    evt = self._intent_stream_handler(evt, stream, mode)
                elif evt.event_type == 'wait':
                    evt = self._wait_stream_handler(evt, stream, mode)
                elif evt.event_type == 'sys':
                    evt = self._sys_stream_handler(evt, stream, mode)
                elif evt.event_type == 'info':
                    evt = self._info_stream_handler(evt, stream, mode)
                elif evt.event_type == 'result':
                    evt = self._result_stream_handler(evt, stream)
                elif evt.event_type == 'user':
                    evt = self._user_stream_handler(stream)
                else:
                    raise NotImplementedError(evt.event_type)


class RunPipeline(SessionContext, ABC):

    def _submit_prompt(self, prompt: str):
        return self.client.submit_prompt_to_pipeline(prompt, self.session_meta.obj_id.identifier)

    def call(self, prompt, ix=None, total=None):

        msg = 'Pipeline is processing prompt'
        if ix is not None:
            msg += f' ({ix}'
        if total is not None:
            msg += f'/{total})'
        elif ix is not None:
            msg += ')'

        with Spinner(msg) as spinner:
            res = self._submit_prompt(prompt)
            spinner.done = f"Prompt processed ({res['metadata']['request_id']})"
        return res

    def map(self, prompts: Iterable[str]) -> Iterable[dict]:
        total = None
        if isinstance(prompts, Sized):
            total = len(prompts)

        for i, prompt in enumerate(prompts):
            yield self.call(prompt, i+1, total)


class RunAgentApp(RunAgent):

    def __init__(self, client, slot_type, model_id, app_id, tool_module_override):
        self.app_id = app_id
        super().__init__(client, slot_type, model_id, tool_module_override)

    def _start_session(self) -> ObjectMetadata:
        sess = self.client.start_session(self.app_id, self.model_id)
        return sess


class RunPipelineApp(RunPipeline):

    def __init__(self, client, slot_type, model_id, app_id, tool_module_override):
        self.app_id = app_id
        super().__init__(client, slot_type, model_id, tool_module_override)

    def _start_session(self) -> ObjectMetadata:
        sess = self.client.start_session(self.app_id, self.model_id)
        return sess


class RunAgentCircuit(RunAgent):

    def __init__(self, client, slot_type, model_id, directive, circuit, tool_module_override):
        self.directive = directive
        self.circuit = circuit
        super().__init__(client, slot_type, model_id, tool_module_override)

    def __enter__(self):
        return super().__enter__()

    def _start_session(self) -> ObjectMetadata:
        return self.client.start_dev_session(self.circuit, self.directive, self.model_id, tool_module_override=self.tool_module_override)


class RunPipelineCircuit(RunPipeline):

    def __init__(self, client, slot_type, model_id, directive, circuit, tool_module_override):
        self.directive = directive
        self.circuit = circuit
        super().__init__(client, slot_type, model_id, tool_module_override)

    def __enter__(self):
        return super().__enter__()

    def _start_session(self) -> ObjectMetadata:
        return self.client.start_dev_session(self.circuit, self.directive, self.model_id, tool_module_override=self.tool_module_override)
